# Copyright 2018 Yongjin Cho
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import sys
import csv
import random
import logging
import multiprocessing as mp


class Dataset:
    def __iter__(self):
        return DatasetIterator(self)

    def samples(self):
        raise NotImplemented

    def map(self, map_fn, workers=None):
        return MappedDataset(self, map_fn)

    def filter(self, filter_fn):
        return FilteredDataset(self, filter_fn)

    def repeat(self, count=None):
        return RepeatedDataset(self, count)

    def shuffle(self, buffer_size):
        return ShuffledDataset(self, buffer_size)

    def batch(self, batch_size, collate_fn):
        return BatchedDataset(self, batch_size, collate_fn)

    def bucket(self, boundaries, batch_sizes, length_fn, collate_fn):
        return BucketDataset(self, boundaries, batch_sizes, length_fn, collate_fn)


class DatasetIterator:
    def __init__(self, dataset):
        self.dataset = dataset
        self.q = mp.Queue(10)
        self.p = mp.Process(target=self._fill_queue, daemon=True)
        self.p.start()

    def __del__(self):
        self.p.terminate()
        self.p.join()

    def __next__(self):
        sample = self.q.get()
        if sample is None:
            raise StopIteration
        return sample

    def _fill_queue(self):
        for sample in self.dataset.samples():
            self.q.put(sample)
        self.q.put(None)


class MappedDataset(Dataset):
    def __init__(self, source, map_fn):
        self.source = source
        self.map_fn = map_fn

    def samples(self):
        for sample in self.source.samples():
            yield self.map_fn(sample)


class FilteredDataset(Dataset):
    def __init__(self, source, filter_fn):
        self.source = source
        self.filter_fn = filter_fn

    def samples(self):
        for sample in self.source.samples():
            if self.filter_fn(sample):
                yield sample


class RepeatedDataset(Dataset):
    def __init__(self, source, count=None):
        self.source = source
        self.count = count

    def samples(self):
        c = 0
        while self.count is None or c < self.count:
            for sample in self.source.samples():
                yield sample
            c += 1


class ShuffledDataset(Dataset):
    def __init__(self, source, buffer_size):
        self.source = source
        self.buffer_size = buffer_size

    def samples(self):
        buf = []
        for sample in self.source.samples():
            buf.append(sample)
            if len(buf) == self.buffer_size:
                random.shuffle(buf)
                for sample in buf:
                    yield sample
                buf = []

        if buf:
            for sample in buf:
                yield sample


class BatchedDataset(Dataset):
    def __init__(self, source, batch_size, collate_fn):
        self.source = source
        self.batch_size = batch_size
        self.collate_fn = collate_fn

    def samples(self):
        batch = []
        for sample in self.source.samples():
            batch.append(sample)
            if len(batch) == self.batch_size:
                yield self.collate_fn(batch)
                batch = []

        if batch:
            yield self.collate_fn(batch)


class BucketDataset(Dataset):
    """Tensor2Tensor's bucketing method"""

    def __init__(self, source, boundaries, batch_sizes, length_fn, collate_fn):
        self.source = source
        self.boundaries = boundaries
        self.batch_sizes = batch_sizes
        self.length_fn = length_fn
        self.collate_fn = collate_fn

    def samples(self):
        buckets = [[] for i in range(len(self.boundaries))]
        for sample in self.source.samples():
            length = self.length_fn(sample)
            for i, boundary in enumerate(self.boundaries):
                if length <= boundary:
                    buckets[i].append(sample)
                    if len(buckets[i]) == self.batch_sizes[i]:
                        yield self.collate_fn(buckets[i])
                        buckets[i] = []
                    break

        for bucket in buckets:
            if bucket:
                yield self.collate_fn(bucket)


class TextLineDataset(Dataset):
    def __init__(self, filename):
        if not filename:
            logging.info("'stdin' will be used for dataset because a file wasn't given.")
        self.filename = filename
        self.newline_pattern = re.compile(r'\r?\n$')

    def samples(self):
        if not self.filename:
            f = sys.stdin
        else:
            f = open(self.filename)
        for line in f:
            yield self.newline_pattern.sub('', line)


class CsvDataset(Dataset):
    def __init__(self, filename):
        self.filename = filename

    def itmes(self):
        reader = csv.DictReader(open(self.filename))
        for row in reader:
            yield row
